/*
 * Copyright (c) 2023-2024 VMware Carbon Black, all rights reserved
 */

// Package functional -- functional tests for use with a pipeline or partial environment
//   - we generally don't want this to run outside the partial environment or pipeline in order to prevent pollution to a live data stream
package functional

import (
	"context"
	"fmt"
	"math/rand"
	"os"
	"reflect"
	"strconv"
	"strings"
	"testing"
	"time"

	"carbonblack.com/protocols/psc/devicelastcontacttime"
	"github.com/aws/aws-sdk-go/aws"
	"github.com/aws/aws-sdk-go/aws/credentials"
	"github.com/aws/aws-sdk-go/aws/session"
	"github.com/aws/aws-sdk-go/service/kinesis"
	deagg "github.com/awslabs/kinesis-aggregation/go/deaggregator"
	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/logger"
	"gitlab.bit9.local/psc/platform/ems/hello/last-contact-updater/internal/publisher"
	"gitlab.bit9.local/psc/platform/orgkey-go/v3"
	"google.golang.org/protobuf/proto"
	elasticV6 "gopkg.in/olivere/elastic.v6"
)

const (
	numDevices = 20
)

//goland:noinspection HttpUrlsUsage
var (
	// Information grabbed from test environment
	awsEndpoint     = os.Getenv("AWS_ENDPOINT")
	awsRegion       = os.Getenv("REGION")
	lctStreamName   = os.Getenv("LCT_STREAM_NAME")
	assetStreamName = os.Getenv("ASSET_REGISTRY_STREAM_NAME")
	esMasterNodes   = os.Getenv("ES_MASTER_NODES")
	esNode          = fmt.Sprintf("http://%s", strings.Split(esMasterNodes, "@")[1])

	pipelineMode = os.Getenv("PIPELINE_MODE")

	// Used for device generation
	baseDeviceID = rand.Int63() //nolint:gosec  // we don't need crypto safe rand for tests
	orgID        = rand.Int()   //nolint:gosec  // we don't need crypto safe rand for tests
)

// getKinesisClient -- get a new kinesis client based on your region
func getKinesisClient() (client *kinesis.Kinesis, err error) {
	// Define configuration for aws based on our region
	awsConfig := aws.NewConfig().WithRegion(awsRegion)

	staticCreds := credentials.NewStaticCredentials("test", "test", "")
	awsConfig = awsConfig.WithEndpoint(awsEndpoint).WithCredentials(staticCreds)

	sess, err := session.NewSession(awsConfig)
	if err != nil {
		return
	}

	return kinesis.New(sess), nil
}

var _ = Describe("Last Contact Updater Functional Tests", Ordered, func() {
	inPipeline := false
	if pipelineMode != "" {
		inPipeline = true
	}

	now := time.Now().UnixMilli()
	orgKey, _ := orgkey.Encode("local", orgID)
	orgIDStr := strconv.Itoa(orgKey.OrgID()) // taken from LCT code

	var esClient *elasticV6.Client
	var client *kinesis.Kinesis
	var err error

	BeforeAll(func() {
		hf := logger.HeaderFunc("LCT Functional Test Setup")
		defer hf()

		_, _ = fmt.Fprintf(GinkgoWriter, "----------------------------------------------------------------------\n")
		_, _ = fmt.Fprintf(GinkgoWriter, "AWS_ENDPOINT: %s\n", awsEndpoint)
		Expect(awsEndpoint).ToNot(BeEquivalentTo(""))
		_, _ = fmt.Fprintf(GinkgoWriter, "REGION: %s\n", awsRegion)
		Expect(awsEndpoint).ToNot(BeEquivalentTo(""))
		_, _ = fmt.Fprintf(GinkgoWriter, "LCT_STREAM_NAME: %s\n", lctStreamName)
		Expect(lctStreamName).ToNot(BeEquivalentTo(""))
		_, _ = fmt.Fprintf(GinkgoWriter, "ASSET_REGISTRY_STREAM_NAME: %s\n", assetStreamName)
		Expect(assetStreamName).ToNot(BeEquivalentTo(""))
		_, _ = fmt.Fprintf(GinkgoWriter, "ES_MASTER_NODES: %s\n", esMasterNodes)
		Expect(esNode).ToNot(BeEquivalentTo(""))

		_, _ = fmt.Fprintf(GinkgoWriter, "OrgKey: %s, OrgId: %d\n", orgKey.String(), orgID)
		_, _ = fmt.Fprintf(GinkgoWriter, "Using Now: %d\n", now)
		_, _ = fmt.Fprintf(GinkgoWriter, "----------------------------------------------------------------------\n\n")

		_, _ = fmt.Fprintf(GinkgoWriter, "Defining kinesis client and checking streams ...")
		client, err = getKinesisClient()
		if err != nil {
			_, _ = fmt.Fprintf(GinkgoWriter, "\n### Failed: %v", err)
		}
		Expect(err).ShouldNot(HaveOccurred())
		Expect(client.WaitUntilStreamExists(&kinesis.DescribeStreamInput{StreamName: aws.String(lctStreamName)})).To(BeNil())
		Expect(client.WaitUntilStreamExists(&kinesis.DescribeStreamInput{StreamName: aws.String(assetStreamName)})).To(BeNil())
		_, _ = fmt.Fprintf(GinkgoWriter, " DONE!\n")

		_, _ = fmt.Fprintf(GinkgoWriter, "Defining elasticsearch client ...")
		esClient, err = elasticV6.NewClient(elasticV6.SetSniff(false), elasticV6.SetURL(esNode), elasticV6.SetScheme("http"))
		if err != nil {
			_, _ = fmt.Fprintf(GinkgoWriter, "\n### Failed: %v", err)
		}
		Expect(err).ShouldNot(HaveOccurred())

		dat, err := os.ReadFile("device_settings.json")
		if err != nil {
			_, _ = fmt.Fprintf(GinkgoWriter, "\n### Unable to read device_settings.json: %v", err)
		}
		Expect(err).ShouldNot(HaveOccurred())
		createIndexResult, err2 := esClient.CreateIndex("a_devicestats").BodyString(string(dat)).Do(context.Background())

		if err2 != nil {
			_, _ = fmt.Fprintf(GinkgoWriter, "\n### Failed to CreateIndex: %v", err2)
		} else {
			_, _ = fmt.Fprintf(GinkgoWriter, " CreateIndex DONE: %+v\n", createIndexResult)
		}
		Expect(err).ShouldNot(HaveOccurred())

		_, _ = fmt.Fprintf(GinkgoWriter, "  -----\n")

		// Create required number of devices for the lct stream and make sure docs exist for each device in ES
		records := make([]*kinesis.PutRecordsRequestEntry, numDevices)
		for i := 0; i < numDevices; i++ {
			currentDeviceID := int64(i) + baseDeviceID
			// Create "original" for elasticsearch index
			devicesStub := publisher.DeviceStatsStub{
				OrganizationID:      int64(orgID),
				DeviceID:            currentDeviceID,
				LastContact:         1000, // non-zero to check for differences
				DeregisterAfterTime: 2000, // non-zero to check for differences
			}

			// Update elasticsearch with device stub
			deviceDocCreateResponse, err2 := esClient.Index().
				Index("a_devicestats").Type("device").
				Id(fmt.Sprintf("%s-%s", orgIDStr, strconv.FormatInt(currentDeviceID, 10))).
				Routing(orgIDStr).BodyJson(devicesStub).Refresh("true").
				Do(context.Background())
			Expect(err2).ShouldNot(HaveOccurred())
			_, _ = fmt.Fprintf(GinkgoWriter, "Pushed to ES: %+v\n   Response: %+v\n", devicesStub, deviceDocCreateResponse)

			// Create device protobuf object for lct kinesis for this device
			lctDevice := &devicelastcontacttime.DeviceLastContactTime{
				OrgKey:               orgKey.String(),
				DeviceId:             currentDeviceID,
				LastContactTimestamp: now, // expected to be new value
				DeregisterAfterTime:  now, // expected to be zeroed out
			}

			// Marshal device and add to the record list for kinesis.PutRecords
			protobufBytes, _ := proto.Marshal(lctDevice) // we expect no problems for known good data
			records[i] = &kinesis.PutRecordsRequestEntry{
				Data:         protobufBytes,
				PartitionKey: aws.String(orgKey.String()),
			}
		}
		esClient.Refresh("a_devicestats")

		_, _ = fmt.Fprintf(GinkgoWriter, "  -----\n")

		// Push the device records onto the LCT stream for access by the service
		_, err = client.PutRecords(&kinesis.PutRecordsInput{
			Records:    records,
			StreamName: aws.String(lctStreamName),
		})
		Expect(err).ShouldNot(HaveOccurred())
		_, _ = fmt.Fprintf(GinkgoWriter, " DONE!\n")

		if inPipeline {
			_, _ = fmt.Fprintf(GinkgoWriter, "Sleeping to allow kinesis data to propegate ...\n")
			time.Sleep(time.Second * 10)
		}
	})

	// ----- Begin actual tests --------------------------------------------------

	Context("Verify LCT Messages", func() {
		It("Check updates to asset registry stream", func() {
			hf := logger.HeaderFunc("Verify Asset Stream Updates")
			defer hf()

			shards, err := client.ListShards(&kinesis.ListShardsInput{StreamName: aws.String(assetStreamName)})
			if err != nil {
				_, _ = fmt.Fprintf(GinkgoWriter, "### Failed to ListShards: %+v\n", err)
			} else {
				_, _ = fmt.Fprintf(GinkgoWriter, "Listed Shards: %+v\n", shards)
			}
			Expect(err).ShouldNot(HaveOccurred())

			shardIteratorOutput, err := client.GetShardIterator(&kinesis.GetShardIteratorInput{ShardId: shards.Shards[0].ShardId, StreamName: aws.String(assetStreamName), ShardIteratorType: aws.String("TRIM_HORIZON")})
			if err != nil {
				_, _ = fmt.Fprintf(GinkgoWriter, "### Failed to GetShardIterator: %+v\n", err)
			} else {
				_, _ = fmt.Fprintf(GinkgoWriter, "GetShardIterator: %+v\n", shardIteratorOutput)
			}
			Expect(err).ShouldNot(HaveOccurred())

			assetRecords, err := client.GetRecords(&kinesis.GetRecordsInput{ShardIterator: shardIteratorOutput.ShardIterator})
			if err != nil {
				_, _ = fmt.Fprintf(GinkgoWriter, "### Failed to GetRecords: %+v\n", err)
			}
			Expect(err).ShouldNot(HaveOccurred())

			_, _ = fmt.Fprintf(GinkgoWriter, "Found %d records:\n%+v\n", len(assetRecords.Records), assetRecords.Records)
			Expect(len(assetRecords.Records)).ToNot(BeEquivalentTo(0))

			sanitized, err2 := deagg.DeaggregateRecords(assetRecords.Records)
			if err2 != nil {
				_, _ = fmt.Fprintf(GinkgoWriter, "Failed to deaggregate records:: %v\n", err)
			}
			Expect(err2).ShouldNot(HaveOccurred())
			for _, v := range sanitized {
				recList := &devicelastcontacttime.DeviceLastContactTime{}
				err := proto.Unmarshal(v.Data, recList)
				if err != nil {
					_, _ = fmt.Fprintf(GinkgoWriter, "Unparsed raw: %s\n  %v\n", string(v.Data), err)
				} else {
					_, _ = fmt.Fprintf(GinkgoWriter, "Data: %+v\n", recList)
				}
			}
			Expect(len(sanitized)).To(BeEquivalentTo(20)) // expect all 20 devices
		})

		It("Check updates to elasticsearch", func() {
			hf := logger.HeaderFunc("Verify Elasticsearch Updates")
			defer hf()

			idRangeQuery := elasticV6.NewRangeQuery("deviceId").Gte(fmt.Sprintf("%d", baseDeviceID)).Lte(fmt.Sprintf("%d", baseDeviceID+numDevices))
			_, _ = fmt.Fprintf(GinkgoWriter, "idRangeQuery: %+v\n", idRangeQuery)
			orgIDQuery := elasticV6.NewTermQuery("organizationId", orgID)
			_, _ = fmt.Fprintf(GinkgoWriter, "orgIDQuery: %+v\n", orgIDQuery)
			fullQuery := elasticV6.NewBoolQuery().Must(orgIDQuery, idRangeQuery)

			searchResult, err := esClient.Search().
				Index("a_devicestats").
				Query(fullQuery).
				Sort("deviceId", true).
				Size(numDevices).
				Do(context.Background())
			if err != nil {
				_, _ = fmt.Fprintf(GinkgoWriter, "### Failed to Do query: %+v\n", err)
			} else {
				_, _ = fmt.Fprintf(GinkgoWriter, "Results: %+v\n", searchResult)
			}
			Expect(err).ShouldNot(HaveOccurred())
			Expect(len(searchResult.Hits.Hits)).ToNot(BeEquivalentTo(0))

			// check each entry's last contact time
			var deviceStub publisher.DeviceStatsStub
			got := 0
			for _, hit := range searchResult.Each(reflect.TypeOf(deviceStub)) {
				_, _ = fmt.Fprintf(GinkgoWriter, "Hit: %+v\n", hit)
				hitAsDeviceStub, ok := hit.(publisher.DeviceStatsStub)
				got++

				Expect(ok).To(BeEquivalentTo(true))
				Expect(hitAsDeviceStub.OrganizationID).To(BeEquivalentTo(int64(orgID))) // should be a part of this test
				Expect(hitAsDeviceStub.LastContact).To(BeEquivalentTo(now))             // should be updated
				Expect(hitAsDeviceStub.DeregisterAfterTime).To(BeEquivalentTo(0))       // should be cleared
			}
			Expect(got).To(BeEquivalentTo(numDevices))
		})
	})
})

func TestLCTApp(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "LCT Functional Suite")
}
